import logging
from threading import Thread

from language_pipes.oai_server import OAIHttpServer

from language_pipes.jobs.job_factory import JobFactory
from language_pipes.jobs.job_receiver import JobReceiver
from language_pipes.jobs.job_tracker import JobTracker

from language_pipes.pipes.router_pipes import RouterPipes
from language_pipes.pipes.pipe_manager import PipeManager

from language_pipes.modeling.model_manager import ModelManager

from language_pipes.util import stop_thread
from language_pipes.config import LpConfig
from language_pipes.network_protocol import StateNetworkNode

class LanguagePipes:
    logger: logging.Logger
    router: StateNetworkNode
    
    job_factory: JobFactory
    job_receiver: JobReceiver

    oai_server: OAIHttpServer
    oai_thread: Thread
    
    config: LpConfig

    def __init__(
        self, 
        config: LpConfig,
        router: StateNetworkNode
    ):
        self.config = config
        router.set_receive_cb(self.receive_data)
        router.set_update_cb(self.print_pipes)
        router.set_disconnect_cb(self.print_pipes)

        self.logger = logging.getLogger(f"LP-{self.config.node_id}")
        self.set_logging_level(self.config.logging_level)
        
        self.router_pipes = None
        self.router = router

        # Network pipe data for MetaPipe objects
        self.router_pipes = RouterPipes(router)

        # Local pipe data for LlmModel objects
        self.model_manager = ModelManager(
            logger=self.logger,
            config=self.config,
            # Used for placing model data on the network
            router_pipes=self.router_pipes
        )

        # Merge local and network data to get Pipe object
        self.pipe_manager = PipeManager(
            config=self.config,
            model_manager=self.model_manager,
            router_pipes=self.router_pipes
        )

        # View currently loaded pipes
        self.router_pipes.print_pipes(self.logger)

        # Holds pending jobs
        self.job_tracker = JobTracker(self.logger, self.config)

        # Handles job creation
        self.job_factory = JobFactory(
            logger=self.logger,
            config=self.config, 
            job_tracker=self.job_tracker,
            pipe_manager=self.pipe_manager
        )

        # Receives jobs and creates JobProcessor object before processing
        self.job_receiver = JobReceiver(
            logger=self.logger, 
            config=self.config, 
            job_tracker=self.job_tracker,
            job_factory=self.job_factory,
            pipe_manager=self.pipe_manager,
            model_manager=self.model_manager,
            is_shutdown=router.is_shut_down
        )

        if self.config.oai_port is not None:
            self.start_oai()

    def receive_data(self, data: bytes):
        self.job_receiver.receive_data(data)

    def print_pipes(self):
        if self.router_pipes is None:
            return
        self.router_pipes.print_pipes(self.logger)

    def start_oai(self):
        if self.config.oai_port is None:
            self.logger.error("Tried to start Open AI server but no port was specified")
            return
        self.oai_server = OAIHttpServer(self.config.oai_port, self.job_factory.start_job, self.router_pipes.get_models)
        self.oai_thread = Thread(target=self.oai_server.serve_forever, args=())
        self.oai_thread.start()
        self.job_factory.logger.info(f"OpenAI Server started on port {self.config.oai_port}")

    def set_logging_level(self, logging_level: str):
        level = getattr(logging, logging_level.upper(), None)
        if level is None:
            raise ValueError(f"Invalid logging level: {logging_level}")
        logging.basicConfig(level=level)

    def stop(self):
        self.model_manager.stop()
        self.router.stop()
        if self.config.oai_port is not None:
            self.oai_server.shutdown()
            stop_thread(self.oai_thread)



from typing import List, Optional, Callable

from promise import Promise

from language_pipes.util import raise_exception
from language_pipes.util.chat import ChatMessage

from language_pipes.pipes.pipe_manager import PipeManager

from language_pipes.jobs.job import Job
from language_pipes.jobs.job_tracker import JobTracker

from language_pipes.config import LpConfig

from language_pipes.modeling.llm_meta_data import validate_model

class JobFactory:
    config: LpConfig
    job_tracker: JobTracker
    pipe_manager: PipeManager

    def __init__(
        self, 
        logger,
        config: LpConfig,
        job_tracker: JobTracker,
        pipe_manager: PipeManager
    ):
        self.config = config
        self.logger = logger
        self.job_tracker = job_tracker
        self.pipe_manager = pipe_manager

    def start_job(
        self, 
        model_id: str, 
        messages: List[ChatMessage], 
        max_completion_tokens: int, 
        temperature: float = 1.0,
        top_k: int = 0,
        top_p: float = 1.0,
        min_p: float = 0.0,
        presence_penalty: float = 0.0,
        start: Optional[Callable] = None,
        update: Optional[Callable] = None,
        resolve: Optional[Promise] = None
    ) -> Optional[Job]:
        pipe = self.pipe_manager.get_pipe_by_model_id(model_id)
        if pipe is None:
            resolve('No pipe available')
            raise_exception(self.logger, f"Could not find pipe for model {model_id}")
            return

        job = Job(
            origin_node_id=self.config.node_id,
            messages=messages, 
            pipe_id=pipe.pipe_id, 
            model_id=pipe.model_id,
            temperature=temperature, 
            prefill_chunk_size=self.config.prefill_chunk_size,
            top_k=top_k, 
            top_p=top_p, 
            min_p=min_p, 
            presence_penalty=presence_penalty,
            max_completion_tokens=max_completion_tokens,
            resolve=resolve,
            update=update,
            complete=self.job_tracker.complete_job
        )
        
        job.print_job(self.logger)
        
        network_job = job.to_network_job()
        pipe.send_job(network_job, self.config.node_id)
        self.job_tracker.jobs_pending.append(job)

        if start is not None:
            start(job)

        return job



import json
import ctypes
import struct
import threading
from hashlib import sha256

def int_to_bytes(i: int) -> bytes:
    return i.to_bytes(4, 'little', signed=False)

def bytes_to_int(b: bytes) -> int:
    return int.from_bytes(b, 'little', signed=False)

def float_to_bytes(f: float) -> bytes:
    return struct.pack(">d", f)

def bytes_to_float(b: bytes) -> float:
    return struct.unpack(">d", b)[0]

def get_byte_hash(data: bytes) -> bytes:
    return sha256(data).digest()

def get_hash(data: str) -> bytes:
    return get_byte_hash(data.encode('utf-8'))

def get_dict_hash(data: dict) -> bytes:
    return get_hash(json.dumps(data))

def stop_thread(thread: threading.Thread):
    thread_id = thread.ident
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id,
            ctypes.py_object(SystemExit))
    if res > 1: # pragma: no cover
        ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
        print('Exception raise failure')


from typing import Dict, Optional

from distributed_state_network.objects.endpoint import Endpoint
from distributed_state_network.util.byte_helper import ByteHelper
from distributed_state_network.objects.signed_packet import SignedPacket

class PeersPacket(SignedPacket):
    node_id: str
    connections: Dict[str, Endpoint]

    def __init__(self, node_id: str, ecdsa_signature: Optional[bytes], connections: Dict[str, Endpoint]):
        super().__init__(ecdsa_signature)
        self.node_id = node_id
        self.connections = connections

    def to_bytes(self, include_signature: bool = True) -> bytes:
        bts = ByteHelper()
        bts.write_string(self.node_id)
        if include_signature :
            if self.ecdsa_signature is None:
                raise ValueError("Cannot convert unsigned packet to bytes")
            bts.write_bytes(self.ecdsa_signature)
        
        bts.write_int(len(self.connections.keys()))
        for key in self.connections.keys():
            bts.write_string(key)
            bts.write_bytes(self.connections[key].to_bytes())
        
        return bts.get_bytes()

    @staticmethod
    def from_bytes(data: bytes):
        bts = ByteHelper(data)
        node_id = bts.read_string()
        ecdsa_signature = bts.read_bytes()

        connections = { }
        num_keys = bts.read_int()
        for _ in range(num_keys):
            key = bts.read_string()
            connections[key] = Endpoint.from_bytes(bts.read_bytes())
        
        return PeersPacket(node_id, ecdsa_signature, connections)



import os
import sys
import unittest

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', 'src'))

from language_pipes.modeling.llm_meta_data import LlmMetadata
from language_pipes.modeling.meta_model import MetaModel
from language_pipes.pipes.meta_pipe import MetaPipe
from language_pipes.pipes.router_pipes import aggregate_models

def make_computed():
    metadata = LlmMetadata()
    metadata.embed_size = 128
    metadata.head_size = 256
    metadata.avg_layer_size = 64
    metadata.embed_hash = "embed"
    metadata.head_hash = "head"
    metadata.layer_hashes=["l0", "l1", "l2", "l3"]
    return metadata


class MetaPipeTests(unittest.TestCase):
    def test_is_complete_with_contiguous_loaded_segments(self):
        meta_data = make_computed()
        segments = [
            MetaModel("p1", 0, 1, True, "node-a", "pipe-1", "model-1", 4, meta_data),
            MetaModel("p2", 2, 3, True, "node-b", "pipe-1", "model-1", 4, meta_data),
        ]
        pipe = MetaPipe("pipe-1", "model-1", segments)

        self.assertTrue(pipe.is_complete())

    def test_is_incomplete_with_gap(self):
        meta_data = make_computed()
        segments = [
            MetaModel("p1", 0, 1, True, "node-a", "pipe-1", "model-1", 4, meta_data),
            MetaModel("p2", 3, 3, True, "node-b", "pipe-1", "model-1", 4, meta_data),
        ]
        pipe = MetaPipe("pipe-1", "model-1", segments)

        self.assertFalse(pipe.is_complete())

    def test_next_start_and_end_layer(self):
        meta_data = make_computed()
        segments = [
            MetaModel("p1", 0, 1, True, "node-a", "pipe-1", "model-1", 4, meta_data),
        ]
        pipe = MetaPipe("pipe-1", "model-1", segments)

        self.assertEqual(pipe.next_start_layer(), 2)
        self.assertEqual(pipe.next_end_layer(), 3)

    def test_aggregate_models_sorts_segments(self):
        meta_data = make_computed()
        models = [
            MetaModel("p2", 2, 3, True, "node-b", "pipe-1", "model-1", 4, meta_data),
            MetaModel("p1", 0, 1, True, "node-a", "pipe-1", "model-1", 4, meta_data),
        ]
        pipes = aggregate_models(models)

        self.assertEqual(len(pipes), 1)
        self.assertEqual([s.start_layer for s in pipes[0].segments], [0, 2])

    def test_get_filled_slots_marks_loaded_and_loading(self):
        meta_data = make_computed()
        segments = [
            MetaModel("p1", 0, 1, False, "node-a", "pipe-1", "model-1", 4, meta_data),
            MetaModel("p2", 2, 3, True, "node-b", "pipe-1", "model-1", 4, meta_data),
        ]
        pipe = MetaPipe("pipe-1", "model-1", segments)

        self.assertEqual(pipe.get_filled_slots(), [1, 1, 2, 2])


if __name__ == '__main__':
    unittest.main()



from typing import Callable, List, Optional
from pathlib import Path
from uuid import uuid4

from transformers import AutoTokenizer
from transformers.models.auto import AutoConfig
from language_pipes.network_protocol import StateNetworkNode

from language_pipes.pipes.meta_pipe import MetaPipe
from language_pipes.modeling.llm_model import LlmModel
from language_pipes.jobs.network_job import NetworkJob
from language_pipes.util.chat import ChatMessage

class Pipe:
    pipe_id: str
    model_id: str
    segments: List[LlmModel]

    router: StateNetworkNode
    tokenizer: Callable
    
    def __init__(
            self, 
            router: StateNetworkNode,
            pipe_id: Optional[str],
            model_id: str,
            model_dir: str
        ):
        self.router = router
        self.model_id = model_id
        model_path = str(Path(model_dir) / model_id / 'data')
        
        if pipe_id is None:
            self.pipe_id = str(uuid4())
        else:
            self.pipe_id = pipe_id

        self.segments = []
        self.tokenizer = lambda: AutoTokenizer.from_pretrained(model_path)

    def send_job(self, job: NetworkJob, node_id: str):
        data = job.to_bytes()
        if node_id == self.router.node_id():
            self.router._receive_data(data)
        else:
            self.router.send_to_node(node_id, data)

    def tokenize(self, prompt: Optional[str], messages: List[ChatMessage]) -> List[int]:
        tokenizer: AutoTokenizer = self.tokenizer()
        if prompt is None:
            prompt = tokenizer.apply_chat_template([m.to_json() for m in messages], tokenize=False, add_generation_prompt=True, chat_template=tokenizer.chat_template)
        return [int(t) for t in tokenizer.encode(prompt, return_tensors='pt')[0].numpy()]

    def get_layer(self, layer: int, need_physical: bool = False) -> Optional[LlmModel]:
        for segment in self.segments:
            if segment.start_layer == layer and (not need_physical or not segment.virtual):
                return segment
        return None
    
    def num_hidden_layers(self) -> Optional[int]:
        for segment in self.segments:
            if not segment.virtual:
                return segment.num_hidden_layers
        return None

    def get_computed(self):
        return self.segments[0].meta_data

    def sort_segments(self):
        self.segments = sorted(self.segments, key=lambda x: x.start_layer)

    def is_complete(self):
        if len(self.segments) == 0:
            return False
        
        self.sort_segments()
        current_layer = 0
        for s in self.segments:
            if not s.loaded:
                break
            if s.start_layer == current_layer:
                current_layer = s.end_layer + 1

        return current_layer == self.num_hidden_layers()

    @staticmethod
    def from_meta(
        meta_pipe: MetaPipe, 
        hosted_models: List[LlmModel], 
        router: StateNetworkNode,
        model_dir: str
    ) -> 'Pipe':
        p = Pipe(
            model_id=meta_pipe.model_id, 
            pipe_id=meta_pipe.pipe_id,
            model_dir=model_dir,
            router=router
        )
        local_segments = []
        for model in hosted_models:
            if model.pipe_id == meta_pipe.pipe_id:
                p.segments.append(model)
                local_segments.append(model.process_id)
        p.segments.extend([LlmModel.from_meta(s, model_dir) for s in meta_pipe.segments if s.process_id not in local_segments])
        p.sort_segments()
        return p



from time import time
from language_pipes.util.byte_helper import ByteHelper

class JobTime:
    is_embed: bool
    is_head: bool
    receive_time: float
    send_time: float
    start_layer: int
    end_layer: int
    node_id: str

    def __init__(
        self, 
        node_id: str = "",
        is_embed: bool = False, 
        is_head: bool = False,
        start_layer: int = 0,
        end_layer: int = 0
    ):
        self.node_id = node_id
        self.is_embed = is_embed
        self.is_head = is_head
        self.start_layer = start_layer
        self.end_layer = end_layer
        self.receive_time = time()

    def set_send_time(self):
        self.send_time = time()

    def to_bytes(self) -> bytes:
        bts = ByteHelper()
        bts.write_string(self.node_id)
        bts.write_int(1 if self.is_embed else 0)
        bts.write_int(1 if self.is_head else 0)
        bts.write_float(self.receive_time)
        bts.write_float(self.send_time)
        bts.write_int(self.start_layer)
        bts.write_int(self.end_layer)
        return bts.get_bytes()

    @staticmethod
    def from_bytes(data: bytes):
        bts = ByteHelper(data)
        lt = JobTime()
        lt.node_id = bts.read_string()
        lt.is_embed = bts.read_int() == 1
        lt.is_head = bts.read_int() == 1
        lt.receive_time = bts.read_float()
        lt.send_time = bts.read_float()
        lt.start_layer = bts.read_int()
        lt.end_layer = bts.read_int()
        return lt



import time
import random
import logging
import threading
import requests
from typing import Dict, List, Optional, Callable

from distributed_state_network.objects.endpoint import Endpoint
from distributed_state_network.objects.hello_packet import HelloPacket
from distributed_state_network.objects.peers_packet import PeersPacket
from distributed_state_network.objects.state_packet import StatePacket
from distributed_state_network.objects.data_packet import DataPacket
from distributed_state_network.objects.config import DSNodeConfig

from distributed_state_network.util import get_dict_hash
from distributed_state_network.util.key_manager import CredentialManager
from distributed_state_network.util.aes import aes_encrypt, aes_decrypt

TICK_INTERVAL = 3
HTTP_TIMEOUT = 2  # seconds

# Message type constants (must match handler.py)
MSG_HELLO = 1
MSG_PEERS = 2
MSG_UPDATE = 3
MSG_PING = 4
MSG_DATA = 5

# Map message types to endpoint paths
MSG_TYPE_TO_PATH = {
    MSG_HELLO: '/hello',
    MSG_PEERS: '/peers',
    MSG_UPDATE: '/update',
    MSG_PING: '/ping',
    MSG_DATA: '/data'
}

class DSNode:
    version: str
    config: DSNodeConfig
    address_book: Dict[str, Endpoint]
    node_states: Dict[str, StatePacket]
    shutting_down: bool

    def __init__(
            self, 
            config: DSNodeConfig,
            version: str,
            disconnect_callback: Optional[Callable] = None,
            update_callback: Optional[Callable] = None,
            receive_callback: Optional[Callable] = None
        ):
        self.config = config
        self.version = version
        self.shutting_down = False
        
        self.cred_manager = CredentialManager(config.credential_dir, config.node_id)
        self.cred_manager.generate_keys()
        
        self.node_states = {
            self.config.node_id: StatePacket.create(self.config.node_id, time.time(), self.cred_manager.my_private(), { })
        }

        self.address_book = {
            self.config.node_id: Endpoint(self.config.network_ip, config.port)
        }
        
        self.logger = logging.getLogger("DSN: " + config.node_id)
        self.disconnect_cb = disconnect_callback
        self.update_cb = update_callback
        self.receive_cb = receive_callback
        
        threading.Thread(target=self.network_tick, daemon=True).start()

    def get_aes_key(self) -> Optional[bytes]:
        if self.config.aes_key is None:
            return None
        return bytes.fromhex(self.config.aes_key)

    def write_address_book(self, node_id: str, conn: Endpoint):
        self.logger.info(f"Address set: {node_id} -> {conn.address}:{conn.port}")
        self.address_book[node_id] = conn

    def network_tick(self):
        time.sleep(TICK_INTERVAL)
        if self.shutting_down:
            self.logger.info("Shutting down node")
            return
        self.test_connections()
        self.gossip()
        threading.Thread(target=self.network_tick, daemon=True).start()

    def gossip(self):
        if len(self.address_book.keys()) == 0:
            return
        node_id = random.choice(list(self.address_book.keys()))
        if node_id == self.config.node_id:
            return
        
        self.send_update(node_id)

    def test_connections(self):
        def remove(node_id: str):
            if node_id in self.node_states:
                del self.node_states[node_id]
            if node_id in self.address_book:    
                del self.address_book[node_id]
            self.logger.info(f"PING failed for {node_id}, disconnecting...")
        
        for node_id in self.node_states.copy().keys():
            if node_id not in self.node_states or node_id == self.config.node_id:
                continue
            try:
                if self.shutting_down:
                    return
                self.send_ping(node_id)
            except Exception:
                self.logger.error(f"PING Failure for {node_id}")
                if node_id in self.node_states:  # double check if something has changed since the ping request started
                    remove(node_id)
                    if self.disconnect_cb is not None:
                        self.disconnect_cb()

    def send_http_request(self, endpoint: Endpoint, msg_type: int, payload: bytes, retries: int = 0) -> bytes:
        """Send HTTP request and wait for response"""
        try:
            # Prepend message type to payload
            data = bytes([msg_type]) + payload
            if self.config.aes_key is not None:
                data = self.encrypt_data(data)
            
            # Determine the URL path based on message type
            path = MSG_TYPE_TO_PATH.get(msg_type, '/unknown')
            url = f"http://{endpoint.address}:{endpoint.port}{path}"
            
            # Send HTTP POST request
            response = requests.post(
                url,
                data=data,
                headers={'Content-Type': 'application/octet-stream'},
                timeout=HTTP_TIMEOUT
            )
            
            # Check response status
            if response.status_code == 204:
                # No content - valid for some responses like successful HELLO with no data
                return b''
            elif response.status_code != 200:
                raise Exception(f"HTTP error: {response.status_code}")
            
            response_data = response.content
            # Decrypt the response
            if self.config.aes_key is not None:
                response_data = self.decrypt_data(response_data)
            
            if len(response_data) < 1:
                raise Exception("Empty response")
            
            # First byte is message type
            response_msg_type = response_data[0]
            if response_msg_type != msg_type:
                raise Exception(f"Response message type mismatch: expected {msg_type}, got {response_msg_type}")
            
            # Return the body (everything after the message type byte)
            return response_data[1:]
            
        except requests.exceptions.Timeout:
            if retries < 2:
                time.sleep(0.5)
                return self.send_http_request(endpoint, msg_type, payload, retries + 1)
            else:
                raise Exception(f"HTTP request to {endpoint.to_string()} timed out")
        except requests.exceptions.RequestException as e:
            if retries < 2:
                time.sleep(0.5)
                return self.send_http_request(endpoint, msg_type, payload, retries + 1)
            else:
                raise Exception(f"HTTP request to {endpoint.to_string()} failed: {e}")
        except Exception as e:
            if retries < 2:
                time.sleep(0.5)
                return self.send_http_request(endpoint, msg_type, payload, retries + 1)
            else:
                raise Exception(f"HTTP request to {endpoint.to_string()} failed: {e}")

    def send_request_to_node(self, node_id: str, msg_type: int, payload: bytes) -> bytes:
        con = self.connection_from_node(node_id)
        return self.send_http_request(con, msg_type, payload)

    def encrypt_data(self, data: bytes) -> bytes:
        key = self.get_aes_key()
        if key is None:
            return data
        return aes_encrypt(key, data)

    def decrypt_data(self, data: bytes) -> bytes:
        key = self.get_aes_key()
        if key is None:
            return data
        return aes_decrypt(key, data)

    def request_peers(self, node_id: str):
        pkt = PeersPacket(self.config.node_id, None, { })
        pkt.sign(self.cred_manager.my_private())
        content = self.send_request_to_node(node_id, MSG_PEERS, pkt.to_bytes())
        pkt = PeersPacket.from_bytes(content)
        if not pkt.verify_signature(self.cred_manager.read_public(node_id)):
            raise Exception("Could not verify peers packet")

        for key in pkt.connections.keys():
            if key == self.config.node_id:
                continue
            
            self.write_address_book(key, pkt.connections[key])
            
            if key not in self.node_states:
                self.send_hello(self.address_book[key])
                
            node_state = self.send_update(key)
            self.handle_update(node_state)

    def handle_peers(self, data: bytes):
        pkt = PeersPacket.from_bytes(data)
        if pkt.node_id not in self.address_book:
            raise Exception(401, f"Could not find {pkt.node_id} in address book")  # Not Authorized
        
        if not pkt.verify_signature(self.cred_manager.read_public(pkt.node_id)):
            raise Exception(406, "Could not verify ECDSA signature of packet")  # Not Acceptable

        peers = { }
        for key in self.address_book.keys():
            if key == self.config.node_id:
                continue
            peers[key] = self.address_book[key]
        
        pkt = PeersPacket(self.config.node_id, None, peers)
        pkt.sign(self.cred_manager.my_private())
        return pkt.to_bytes()

    def send_hello(self, con: Endpoint):
        self.logger.info(f"HELLO => {con.to_string()}")

        pkt = self.my_hello_packet()
        payload = pkt.to_bytes()
        content = self.send_http_request(con, MSG_HELLO, payload)
        
        # Get the response packet
        pkt = HelloPacket.from_bytes(content)
        self.logger.info(f"Received HELLO from {pkt.node_id}")
        
        # Verify version compatibility
        if pkt.version != self.version:
            msg = f"HELLO => {pkt.node_id} (Version mismatch \"{pkt.version}\" != \"{self.version}\")"
            self.logger.error(msg)
            raise Exception(505)  # Version not supported

        # Store the peer's public key
        self.cred_manager.ensure_public(pkt.node_id, pkt.ecdsa_public_key)
        
        # If the server sent us our detected IP, update our address book
        if pkt.detected_address:
            self.logger.info(f"Server detected our IP as: {pkt.detected_address}")
            # Update our own connection in the address book with the detected IP
            self.write_address_book(self.config.node_id, Endpoint(pkt.detected_address, self.config.port))
        
        self.write_address_book(pkt.node_id, con)

        if pkt.node_id not in self.node_states:
            self.init_state(pkt.node_id)

        return pkt.node_id

    def init_state(self, node_id: str):
        self.node_states[node_id] = StatePacket(node_id, 0, b'', { })

    def handle_hello(self, data: bytes, detected_address: str) -> bytes:
        pkt = HelloPacket.from_bytes(data)
        self.logger.info(f"Received HELLO from {pkt.node_id}")
        if pkt.version != self.version:
            msg = f"HELLO => {pkt.node_id} (Version mismatch \"{pkt.version}\" != \"{self.version}\")"
            self.logger.error(msg)
            raise Exception(505)  # Version not supported

        self.cred_manager.ensure_public(pkt.node_id, pkt.ecdsa_public_key)
        self.write_address_book(pkt.node_id, Endpoint(detected_address, pkt.connection.port))

        if pkt.node_id not in self.node_states:
            self.init_state(pkt.node_id)

        # Create response with detected address
        response_pkt = self.my_hello_packet()
        response_pkt.detected_address = detected_address
        return response_pkt.to_bytes()

    def my_hello_packet(self) -> HelloPacket:
        pkt = HelloPacket(
            self.version, 
            self.config.node_id, 
            self.my_con(), 
            self.cred_manager.my_public(), 
            None,
            None  # No certificate for HTTP
        )
        pkt.sign(self.cred_manager.my_private())
        return pkt

    def send_ping(self, node_id: str):     
        try:
            conn = self.connection_from_node(node_id)
            self.logger.debug(f"PING => {node_id} ({conn.address}:{conn.port})")
            self.send_request_to_node(node_id, MSG_PING, b' ')
        except Exception as e:
            raise Exception(f'PING => {node_id}: {e}')

    def send_update(self, node_id: str):
        self.logger.debug(f"UPDATE => {node_id}")
        content = self.send_request_to_node(node_id, MSG_UPDATE, self.my_state().to_bytes())
        return content

    def handle_update(self, data: bytes):
        pkt = StatePacket.from_bytes(data)
        
        if not self.update_state(pkt):
            return b''

        self.logger.info(f"Received UPDATE from {pkt.node_id}")

        if self.update_cb is not None:
            try:
                self.update_cb()
            except Exception as e:
                self.logger.error("Update Error Captured:")
                self.logger.error(str(e))

        return self.my_state().to_bytes()

    def my_state(self):
        return self.node_states[self.config.node_id]
    
    def update_state(self, pkt: StatePacket) -> bool:
        # ignore if we accidentally sent an update to ourselves
        if pkt.node_id == self.config.node_id:
            raise Exception(406, "Origin and destination are the same")  # Not acceptable

        if pkt.node_id in self.address_book and not pkt.verify_signature(self.cred_manager.read_public(pkt.node_id)):
            raise Exception(401, "Could not verify ECDSA signature")  # Not authorized

        if pkt.node_id in self.node_states:
            current_state = self.node_states[pkt.node_id]

            # Check if stale
            if current_state.last_update > pkt.last_update:
                return False

            # Check if duplicate packet
            if len(pkt.state_data.keys()) > 0 and get_dict_hash(self.node_states[pkt.node_id].state_data) == get_dict_hash(pkt.state_data):
                return False

        self.node_states[pkt.node_id] = pkt
        return True

    def bootstrap(self, con: Endpoint):
        bootstrap_id = self.send_hello(con)
        content = self.send_update(bootstrap_id)
        self.handle_update(content)
        self.request_peers(bootstrap_id)

    def connection_from_node(self, node_id: str) -> Endpoint:
        if node_id not in self.address_book:
            raise Exception(f"could not find connection for {node_id}")
        return self.address_book[node_id]

    def update_data(self, key: str, val: str):
        self.node_states[self.config.node_id].update_state(key, val, self.cred_manager.my_private())
        for key in list(self.node_states.keys())[:]:
            if key == self.config.node_id:
                continue
            try:
                self.send_update(key)
            except Exception as e:
                print(e)

    def my_con(self) -> Endpoint:
        return self.connection_from_node(self.config.node_id)

    def read_data(self, node_id: str, key: str) -> Optional[str]:
        if key not in self.node_states[node_id].state_data.keys():
            return None
        return self.node_states[node_id].state_data[key]

    def peers(self) -> List[str]:
        return list(self.node_states.keys())

    def send_to_node(self, node_id: str, data: bytes) -> str:
        pkt = DataPacket.create(self.config.node_id, self.cred_manager.my_private(), data)
        response = self.send_request_to_node(node_id, MSG_DATA, pkt.to_bytes())
        try:
            return response.decode('utf-8')
        except Exception:
            return ''
    
    def receive_data(self, data: bytes):
        pkt = DataPacket.from_bytes(data)
        self.logger.info(f"Received DATA from {pkt.node_id} ({len(pkt.data)} bytes)")
        
        # Ensure sender is known
        if pkt.node_id not in self.address_book:
            raise Exception(401, f"Could not find {pkt.node_id} in address book")
        
        # Verify signature using stored public key
        if not pkt.verify_signature(self.cred_manager.read_public(pkt.node_id)):
            raise Exception(401, "Could not verify ECDSA signature of data packet")
        
        if self.receive_cb is not None:
            try:
                self.receive_cb(pkt.data)
            except Exception as e:
                print(e)

        return b'OK'



from time import time
from uuid import uuid4
from typing import List, Optional

import torch
from promise import Promise
from typing import Callable
from transformers.cache_utils import DynamicCache

from language_pipes.jobs.job_data import JobData
from language_pipes.jobs.network_job import NetworkJob
from language_pipes.jobs.timing_stats import TimingStats
from language_pipes.util.chunk_state import ChunkState

from language_pipes.util.enums import ComputeStep, JobStatus
from language_pipes.util.chat import ChatMessage

class Job:
    # IDs
    job_id: str
    pipe_id: str
    model_id: str
    origin_node_id: str
    
    # Computed
    delta: str
    prompt_tokens: int = 0

    # State Info
    input_ids: List[int]
    compute_step: ComputeStep
    status: JobStatus
    current_token: int = 0
    current_layer: int = 0
    data: Optional[JobData]
    messages: List[ChatMessage]
    result: Optional[str]
    last_update: float
    timing_stats: TimingStats
    
    # API params
    top_k: int
    top_p: float
    min_p: float
    temperature: float
    presence_penalty: float
    max_completion_tokens: int

    # Classes
    cache: DynamicCache
    chunking: ChunkState

    # Functions
    resolve: Promise | None
    update: Optional[Callable[["Job"], None]]
    complete: Callable[[], None]

    def __init__(
            self,
            origin_node_id: str,
            messages: List[ChatMessage],
            pipe_id: str,
            model_id: str,
            prefill_chunk_size: int,
            data: Optional[JobData] = None,
            temperature: float = 1.0,
            top_k: int = 0,
            top_p: float = 1.0,
            min_p: float = 0.0,
            presence_penalty: float = 0.0,
            max_completion_tokens: int = 1000,
            resolve: Optional[Promise] = None,
            update: Optional[Callable[["Job"], None]] = None,
            complete: Optional[Callable[["Job"], None]] = None
        ):
        self.pipe_id = pipe_id
        self.model_id = model_id
        self.job_id = str(uuid4())
        self.origin_node_id = origin_node_id
        
        self.status = JobStatus.IN_PROGRESS
        self.compute_step = ComputeStep.TOKENIZE

        self.delta = ''
        self.data = data
        self.result = None
        self.input_ids = []
        self.timing_stats = TimingStats(self.job_id, prefill_chunk_size)
        self.prompt_tokens = 0
        self.current_token = 0
        self.messages = messages

        self.temperature = temperature
        self.top_k = top_k
        self.top_p = top_p
        self.min_p = min_p
        self.presence_penalty = presence_penalty
        self.max_completion_tokens = max_completion_tokens
        
        self.current_layer = 0

        self.cache = DynamicCache()
        self.chunking = ChunkState(self.job_id)
        self.resolve = resolve
        self.update = update

        if complete is not None:
            self.complete = lambda: complete(self)
        else:
            self.complete = self.pass_complete
        
        self.last_update = time()

    def pass_complete(self):
        pass

    def init_chunking(self, chunk_size: int):
        self.chunking.init(self.prompt_tokens, chunk_size)

    def set_layer(self, state: torch.Tensor, layer: int, num_hidden_layers: int):
        if self.compute_step != ComputeStep.LAYER:
            raise Exception('Invalid step for layer')
        self.current_layer = layer
        if self.data is None: 
            return
        self.data.state = state
        if self.current_layer == num_hidden_layers:
            self.compute_step = ComputeStep.EMBED if self.chunking.has_more() else ComputeStep.HEAD
            self.current_layer = 0

    def set_norm(self, state: torch.Tensor):
        if self.compute_step != ComputeStep.NORM:
            raise Exception('Invalid step for norm')
        if self.data is None:
            return
        self.data.state = state
        self.next_step()

    def set_output(self, token: int, eos_token: int):
        if self.compute_step != ComputeStep.HEAD:
            raise Exception('Invalid step for head')
        self.input_ids.append(token)
        self.next_step()
        if token == eos_token:
            self.status = JobStatus.COMPLETED

    def input_id_tensor(self):
        if self.input_ids is None:
            return None
        return torch.tensor(self.input_ids)

    def next_step(self):
        if self.compute_step == ComputeStep.TOKENIZE:
            self.compute_step = ComputeStep.EMBED
        elif self.compute_step == ComputeStep.EMBED:
            self.compute_step = ComputeStep.LAYER
        elif self.compute_step == ComputeStep.LAYER:
            self.compute_step = ComputeStep.NORM
        elif self.compute_step == ComputeStep.NORM:
            self.compute_step = ComputeStep.HEAD
        elif self.current_token < self.max_completion_tokens:
            self.current_token += 1
            self.compute_step = ComputeStep.EMBED
            if self.current_token == self.max_completion_tokens:
                self.status = JobStatus.COMPLETED
        else:
            self.status = JobStatus.COMPLETED

    def receive_network_job(self, network_job: NetworkJob) -> bool:
        if network_job.job_id != self.job_id or network_job.pipe_id != self.pipe_id:
            return False
        if network_job.origin_node_id != self.origin_node_id:
            return False

        if network_job.compute_step == ComputeStep.HEAD and self.chunking.has_more():
            self.compute_step = ComputeStep.EMBED
            self.current_layer = 0
        else:
            self.compute_step = network_job.compute_step
            self.current_layer = network_job.current_layer
            
        self.data = network_job.data
        self.timing_stats.receive_network_job(network_job.times)
        
        return True

    def send_update(self):
        self.last_update_time = time()
        if self.update is not None:
            return self.update(self)
        return True

    def to_network_job(self) -> NetworkJob:
        data_hash = self.data.hash_state() if self.data is not None else b''
        return NetworkJob(
            job_id=self.job_id, 
            pipe_id=self.pipe_id, 
            origin_node_id=self.origin_node_id, 
            current_layer=self.current_layer, 
            data=self.data, 
            data_hash=data_hash, 
            compute_step=self.compute_step, 
            times=list(self.timing_stats.current_times)
        )

    def set_last_update(self):
        from time import time
        self.last_update = time()

    def print_job(self, logger):
        logger.info(f"""
=================================
Job ID: {self.job_id}
Pipe ID: {self.pipe_id}
Prompt Tokens: {self.prompt_tokens}
Current Token: {self.current_token}
Max Tokens: {self.max_completion_tokens}
Temperature: {self.temperature}
Top K: {self.top_k}
Top P: {self.top_p}
Min P: {self.min_p}
Pres Penalty: {self.presence_penalty}
=================================
""")



from typing import Optional
from distributed_state_network.util.ecdsa import verify_signature, sign_message

class SignedPacket:
    ecdsa_signature: Optional[bytes]

    def __init__(self, ecdsa_signature: Optional[bytes]):
        self.ecdsa_signature = ecdsa_signature

    def sign(self, private_key: bytes):
        self.ecdsa_signature = sign_message(private_key, self.to_bytes(False))

    def verify_signature(self, public_key: bytes):
        if self.ecdsa_signature is None:
            raise ValueError("Cannot verifiy, ECDSA signature was not set")
        return verify_signature(public_key, self.to_bytes(False), self.ecdsa_signature)

    def to_bytes(self, include_signature: bool = True) -> bytes:
        return b''

    @staticmethod
    def from_bytes(data: bytes) -> "SignedPacket":
        return SignedPacket(b'')



"""Run all unit tests for language_pipes.

Usage: python -m tests.language_pipes.unit
"""
from . import run_tests
import sys

if __name__ == '__main__':
    success = run_tests()
    sys.exit(0 if success else 1)


from dataclasses import dataclass
from language_pipes.modeling.llm_meta_data import LlmMetadata

@dataclass
class MetaModel:
    process_id: str
    start_layer: int
    end_layer: int
    loaded: bool
    
    node_id: str
    pipe_id: str
    model_id: str
    num_layers: int
    meta_data: LlmMetadata

    def to_json(self):
        return {
            "process_id": self.process_id,
            "start_layer": self.start_layer,
            "end_layer": self.end_layer,
            "node_id": self.node_id,
            "pipe_id": self.pipe_id,
            "model_id": self.model_id,
            "num_layers": self.num_layers,
            "loaded": self.loaded,
            "meta_data": self.meta_data.to_json()
        }

    @staticmethod
    def from_dict(data: dict):
        return MetaModel(
            process_id=data["process_id"],
            start_layer=data["start_layer"],
            end_layer=data["end_layer"],
            loaded=data["loaded"],
            node_id=data["node_id"],
            pipe_id=data["pipe_id"],
            model_id=data["model_id"],
            num_layers=data["num_layers"],
            meta_data=LlmMetadata.from_dict(data["meta_data"])
        )


import time
import json
from typing import Dict

from distributed_state_network.objects.signed_packet import SignedPacket
from distributed_state_network.util.byte_helper import ByteHelper

class StatePacket(SignedPacket):
    node_id: str
    state_data: Dict[str, str]
    last_update: float

    def __init__(
            self, 
            node_id: str,
            last_update: float,
            ecdsa_signature: bytes,
            state_data: Dict[str, str]
        ):
        super().__init__(ecdsa_signature)
        self.node_id = node_id
        self.state_data = state_data
        self.last_update = last_update

    def update_state(self, key: str, val: str, private_key: bytes):
        self.state_data[key] = val
        self.last_update = time.time()
        self.sign(private_key)

    def to_bytes(self, include_signature: bool = True):
        bts = ByteHelper()
        bts.write_string(self.node_id)
        bts.write_float(self.last_update)
        if include_signature:
            bts.write_bytes(self.ecdsa_signature)
        bts.write_string(json.dumps(self.state_data))

        return bts.get_bytes()
    
    @staticmethod
    def create(
        node_id: str, 
        last_update: float,
        ecdsa_private_key: bytes,
        state_data: Dict[str, str]
    ):
        s = StatePacket(node_id, last_update, b'', state_data)
        s.sign(ecdsa_private_key)
        return s


    @staticmethod
    def from_bytes(data: bytes):
        bts = ByteHelper(data)
        node_id = bts.read_string()
        
        if node_id == '':
            raise Exception(406, "Malformed packet")
        
        last_update = bts.read_float()
        ecdsa_signature = bts.read_bytes()
        state_data = json.loads(bts.read_string())

        return StatePacket(node_id, last_update, ecdsa_signature, state_data)



import os
import sys
import torch
import unittest
from time import time

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'src'))
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'tests', 'language_pipes', 'unit'))

from language_pipes.jobs.job_data import JobData
from language_pipes.jobs.job_processor import JobState
from language_pipes.util.enums import ComputeStep, JobStatus

from job_processor.util import mock_complete, make_processor, make_job, make_config, FakeEndModel, FakeModel, FakePipe, FakeEndModelContinue, FakeLogger

class TestHeadState(unittest.TestCase):
    """Tests for the _state_head method."""

    def test_transitions_to_done_on_completion(self):
        completed = []

        def mark_complete(job):
            completed.append(job.job_id)

        job = make_job(complete=mark_complete)
        job.compute_step = ComputeStep.HEAD
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        end_model = FakeEndModel()
        processor = make_processor(job=job, end_model=end_model)

        next_state = processor._state_head()

        self.assertEqual(next_state, JobState.DONE)
        self.assertEqual(job.status, JobStatus.COMPLETED)
        self.assertIn("set_result", end_model.calls)
        self.assertEqual(completed, [job.job_id])

    def test_transitions_to_done_on_update_failure(self):
        updates = []

        def fail_update(job):
            updates.append(job.compute_step)
            return False

        job = make_job(update=fail_update)
        job.compute_step = ComputeStep.HEAD
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        end_model = FakeEndModelContinue()
        processor = make_processor(job=job, end_model=end_model)

        next_state = processor._state_head()

        self.assertEqual(next_state, JobState.DONE)
        self.assertEqual(job.status, JobStatus.IN_PROGRESS)
        self.assertEqual(len(updates), 1)
        self.assertNotIn("set_result", end_model.calls)

    def test_transitions_to_embed_on_successful_update(self):
        updates = []

        def record_update(job):
            updates.append(job.compute_step)
            return True

        job = make_job(update=record_update)
        job.compute_step = ComputeStep.HEAD
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        end_model = FakeEndModelContinue()
        processor = make_processor(job=job, end_model=end_model)

        next_state = processor._state_head()

        self.assertEqual(next_state, JobState.EMBED)
        self.assertEqual(len(updates), 1)
        self.assertIn("compute_norm", end_model.calls)
        self.assertIn("compute_head", end_model.calls)

    def test_origin_mismatch_stops(self):
        job = make_job(origin_node_id="node-b")
        job.compute_step = ComputeStep.HEAD
        job.current_layer = 1
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        processor = make_processor(
            job=job,
            config=make_config(node_id="node-a"),
            end_model=FakeEndModel(),
        )

        processor.run()

        self.assertEqual(processor.state, JobState.DONE)

    def test_logs_job_data_and_timing_when_enabled(self):
        job = make_job(complete=mock_complete)
        job.compute_step = ComputeStep.HEAD
        job.current_layer = 1
        job.prompt_tokens = 2
        job.prefill_start_time = time()
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))
        logger = FakeLogger()

        processor = make_processor(
            job=job,
            config=make_config(),
            logger=logger,
            end_model=FakeEndModel(),
        )

        processor.run()

        self.assertEqual(job.status, JobStatus.COMPLETED)
        self.assertEqual(job.result, "done")
        self.assertTrue(any("Timing" in message for _, message in logger.messages))
        self.assertTrue(any("Job ID" in message for _, message in logger.messages))

class TestHeadFlowIntegration(unittest.TestCase):
    """Integration tests for head state transitions through subsequent states."""

    def test_sends_to_remote_layer(self):
        job = make_job()
        job.compute_step = ComputeStep.HEAD
        job.prompt_tokens = 1
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        end_model = FakeEndModelContinue()
        pipe = FakePipe(FakeModel("node-b", 0, 0, virtual=True))
        processor = make_processor(job=job, pipe=pipe, end_model=end_model)

        head_state = processor._state_head()
        self.assertEqual(head_state, JobState.EMBED)

        next_state = processor._state_embed()
        self.assertEqual(next_state, JobState.SEND)

        final_state = processor._state_send()
        self.assertEqual(final_state, JobState.DONE)
        self.assertEqual(len(pipe.sent_jobs), 1)
        _, node_id = pipe.sent_jobs[0]
        self.assertEqual(node_id, "node-b")

    def test_processes_local_layer(self):
        job = make_job()
        job.compute_step = ComputeStep.HEAD
        job.prompt_tokens = 1
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        end_model = FakeEndModelContinue()
        pipe = FakePipe(FakeModel("node-a", 0, 0, virtual=False))
        processor = make_processor(job=job, pipe=pipe, end_model=end_model)

        head_state = processor._state_head()
        self.assertEqual(head_state, JobState.EMBED)

        next_state = processor._state_embed()
        self.assertEqual(next_state, JobState.PROCESS_LAYERS)



import os
import sys
import unittest

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'src'))
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'tests', 'language_pipes', 'unit'))

from language_pipes.jobs.job_processor import JobState
from language_pipes.util.enums import ComputeStep

from job_processor.util import make_processor, make_job, make_config, FakeEndModel, FakeModel, FakePipe, EmptyPipe

class TestEmbedState(unittest.TestCase):
    """Tests for the _state_embed method."""

    def test_transitions_to_done_when_update_fails(self):
        updates = []

        def fail_update(job):
            updates.append(job.compute_step)
            return False

        job = make_job(update=fail_update)
        job.compute_step = ComputeStep.EMBED
        job.prompt_tokens = 2
        job.current_token = 0
        job.init_chunking(chunk_size=1)

        processor = make_processor(
            job=job,
            config=make_config(prefill_chunk_size=1),
            end_model=FakeEndModel(),
        )

        next_state = processor._state_embed()

        self.assertEqual(next_state, JobState.DONE)
        self.assertEqual(len(updates), 1)

    def test_transitions_to_done_when_model_missing(self):
        job = make_job()
        job.compute_step = ComputeStep.EMBED
        job.prompt_tokens = 1

        processor = make_processor(
            job=job,
            pipe=EmptyPipe(FakeModel("node-a", 0, 0)),
            end_model=FakeEndModel(),
        )

        next_state = processor._state_embed()

        self.assertEqual(next_state, JobState.DONE)

    def test_transitions_to_send_for_remote_layer(self):
        job = make_job()
        job.compute_step = ComputeStep.EMBED
        job.prompt_tokens = 1

        pipe = FakePipe(FakeModel("node-b", 0, 0, virtual=True))
        processor = make_processor(job=job, pipe=pipe, end_model=FakeEndModel())

        next_state = processor._state_embed()

        self.assertEqual(next_state, JobState.SEND)

    def test_transitions_to_process_layers_for_local_layer(self):
        job = make_job()
        job.compute_step = ComputeStep.EMBED
        job.prompt_tokens = 1

        processor = make_processor(
            job=job,
            pipe=FakePipe(FakeModel("node-a", 0, 0, virtual=False)),
            end_model=FakeEndModel(),
        )

        next_state = processor._state_embed()

        self.assertEqual(next_state, JobState.PROCESS_LAYERS)

    def test_transitions_to_process_layers_for_prefill(self):
        job = make_job()
        job.compute_step = ComputeStep.EMBED

        processor = make_processor(
            job=job,
            pipe=FakePipe(FakeModel("node-a", 0, 0, virtual=False)),
            end_model=FakeEndModel(),
        )

        next_state = processor._state_embed()

        self.assertEqual(next_state, JobState.PROCESS_LAYERS)

        job.compute_step = ComputeStep.EMBED
        next_state = processor._state_embed()

        self.assertEqual(next_state, JobState.PROCESS_LAYERS)

class TestEmbedPrefillIntegration(unittest.TestCase):
    """Integration tests for embed state during prefill operations."""

    def test_update_failure_stops(self):
        updates = []
        
        def complete(_):
            pass

        def fail_update(job):
            updates.append(job.compute_step)
            return False

        job = make_job(update=fail_update, complete=complete)
        job.compute_step = ComputeStep.TOKENIZE
        end_model = FakeEndModel()

        processor = make_processor(
            job=job,
            config=make_config(prefill_chunk_size=1),
            end_model=end_model,
        )

        processor.run()

        self.assertEqual(processor.state, JobState.DONE)
        self.assertIn("tokenize", end_model.calls)
        self.assertIn("compute_embed", end_model.calls)

if __name__ == "__main__":
    unittest.main()


from io import BytesIO
from typing import Optional
from distributed_state_network.util import bytes_to_int, int_to_bytes, float_to_bytes, bytes_to_float

class ByteHelper:
    def __init__(self, data: Optional[bytes] = None):
        self.bts = BytesIO(data) if data is not None else BytesIO()

    def write_string(self, s: str):
        encoded = s.encode('utf-8')
        self.write_bytes(encoded)

    def write_int(self, i: int):
        self.bts.write(int_to_bytes(i))

    def write_float(self, f: float):
        self.bts.write(float_to_bytes(f))

    def write_bytes(self, b: bytes):
        self.bts.write(int_to_bytes(len(b)))
        self.bts.write(b)

    def read_string(self):
        return self.read_bytes().decode('utf-8')

    def read_int(self):
        return bytes_to_int(self.bts.read(4))

    def read_float(self):
        return bytes_to_float(self.bts.read(8))

    def read_bytes(self):
        return self.bts.read(bytes_to_int(self.bts.read(4)))

    def get_bytes(self):
        return self.bts.getvalue()



import json
import os
import sys
import unittest

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', 'src'))

from language_pipes.modeling.llm_meta_data import LlmMetadata
from language_pipes.modeling.meta_model import MetaModel
from language_pipes.pipes.router_pipes import RouterPipes


class FakeConnection:
    def __init__(self, address: str):
        self.address = address


class FakeStateNetworkNode:
    def __init__(self, node_id: str):
        self._node_id = node_id
        self._data = {}
        self._peers = []
        self.logger = None

    def node_id(self):
        return self._node_id

    def read_data(self, node_id: str, key: str):
        return self._data.get((node_id, key))

    def update_data(self, key: str, value: str):
        self._data[(self._node_id, key)] = value

    def peers(self):
        return self._peers

    def connection_from_node(self, node_id: str):
        return FakeConnection("127.0.0.1")

    def add_peer(self, peer_id: str, models=None):
        self._peers.append(peer_id)
        if models is None:
            models = []
        self._data[(peer_id, "models")] = json.dumps([m.to_json() for m in models])


def make_computed():
    metadata = LlmMetadata()
    metadata.embed_size = 128
    metadata.head_size = 256
    metadata.avg_layer_size = 64
    metadata.embed_hash = "embed"
    metadata.head_hash = "head"
    metadata.layer_hashes=["l0", "l1", "l2", "l3"]
    return metadata

class RouterPipesTests(unittest.TestCase):
    def test_add_and_get_pipe_by_pipe_id(self):
        node = FakeStateNetworkNode("node-a")
        router = RouterPipes(node)
        meta_data = make_computed()
        model = MetaModel("p1", 0, 1, True, "node-a", "pipe-1", "model-1", 4, meta_data)

        router.add_model_to_network(model)
        node.add_peer("node-a", [model])

        pipe = router.get_pipe_by_pipe_id("pipe-1")

        self.assertIsNotNone(pipe)
        self.assertEqual(pipe.pipe_id, "pipe-1")
        self.assertEqual(pipe.model_id, "model-1")

    def test_pipes_for_model_filters_complete(self):
        node = FakeStateNetworkNode("node-a")
        router = RouterPipes(node)
        meta_data = make_computed()
        node.add_peer(
            "node-a",
            [
                MetaModel("p1", 0, 1, True, "node-a", "pipe-1", "model-1", 4, meta_data),
                MetaModel("p2", 2, 3, True, "node-b", "pipe-1", "model-1", 4, meta_data),
                MetaModel("p3", 0, 1, False, "node-c", "pipe-2", "model-1", 4, meta_data),
            ],
        )

        completed = router.pipes_for_model("model-1", True)
        loading = router.pipes_for_model("model-1", False)

        self.assertEqual([p.pipe_id for p in completed], ["pipe-1"])
        self.assertEqual([p.pipe_id for p in loading], ["pipe-2"])

    def test_update_model_replaces_existing(self):
        node = FakeStateNetworkNode("node-a")
        router = RouterPipes(node)
        meta_data = make_computed()
        model = MetaModel("p1", 0, 1, True, "node-a", "pipe-1", "model-1", 4, meta_data)
        router.add_model_to_network(model)

        updated = MetaModel("p1", 0, 1, False, "node-a", "pipe-1", "model-1", 4, meta_data)
        router.update_model(updated)

        stored = json.loads(node.read_data("node-a", "models"))
        self.assertEqual(stored[0]["loaded"], False)

    def test_get_pipe_by_model_id_returns_complete_pipe(self):
        node = FakeStateNetworkNode("node-a")
        router = RouterPipes(node)
        meta_data = make_computed()
        node.add_peer(
            "node-a",
            [
                MetaModel("p1", 0, 1, True, "node-a", "pipe-1", "model-1", 4, meta_data),
                MetaModel("p2", 2, 3, True, "node-b", "pipe-1", "model-1", 4, meta_data),
            ],
        )

        pipe = router.get_pipe_by_model_id("model-1")

        self.assertIsNotNone(pipe)
        self.assertEqual(pipe.pipe_id, "pipe-1")


if __name__ == "__main__":
    unittest.main()



import toml

def view_config(config_path: str):
    with open(config_path, 'r', encoding='utf-8') as f:
        config = toml.load(f)
    
    print("\n" + "=" * 60)
    print("  Configuration Details")
    print(f"  Path: {config_path}")
    print("=" * 60)
    
    # Property descriptions
    descriptions = {
        "node_id": "Unique identifier for this node on the network",
        "oai_port": "Port for OpenAI-compatible API server",
        "peer_port": "Port for peer-to-peer network coordination",
        "network_ip": "IP address other nodes use to connect to this node",
        "bootstrap_address": "IP of existing node to join the network",
        "bootstrap_port": "Port of the bootstrap node",
        "network_key": "AES encryption key for secure communication",
        "logging_level": "Verbosity level (DEBUG/INFO/WARNING/ERROR)",
        "max_pipes": "Max concurrent distributed model instances",
        "model_validation": "Verify model weight hashes match across nodes",
    }
    
    # Display simple properties
    print("\n--- Node Settings ---")
    _print_property("Node ID", config.get("node_id"), descriptions["node_id"])
    
    print("\n--- Network Settings ---")
    _print_property("Network IP", config.get("network_ip"), descriptions["network_ip"])
    _print_property("Peer Port", config.get("peer_port"), descriptions["peer_port"])
    
    bootstrap = config.get("bootstrap_address")
    if bootstrap:
        _print_property("Bootstrap Address", bootstrap, descriptions["bootstrap_address"])
        _print_property("Bootstrap Port", config.get("bootstrap_port"), descriptions["bootstrap_port"])
    else:
        print("  Bootstrap Node:     Not configured (this is a standalone/first node)")
    
    network_key = config.get("network_key")
    if network_key:
        # Show truncated key for security
        display_key = network_key[:8] + "..." + network_key[-8:] if len(network_key) > 20 else network_key
        _print_property("Network Key", display_key, descriptions["network_key"])
    else:
        print("  Network Encryption: Disabled")
    
    print("\n--- API Settings ---")
    oai_port = config.get("oai_port")
    if oai_port:
        _print_property("OpenAI API Port", oai_port, descriptions["oai_port"])
    else:
        print("  OpenAI API:         Disabled")
    
    print("\n--- Performance & Security ---")
    _print_property("Logging Level", config.get("logging_level", "INFO"), descriptions["logging_level"])
    _print_property("Max Pipes", config.get("max_pipes", 1), descriptions["max_pipes"])
    _print_property("Model Validation", "Enabled" if config.get("model_validation") else "Disabled", descriptions["model_validation"])
    
    # Display hosted models
    models = config.get("hosted_models", [])
    print(f"\n--- Hosted Models ({len(models)}) ---")
    for i, model in enumerate(models):
        print(f"\n  Model #{i+1}:")
        print(f"    ID:          {model.get('id', 'Unknown')}")
        # print(f"                 (HuggingFace model identifier)")
        print(f"    Device:      {model.get('device', 'cpu')}")
        # print(f"                 (Compute device: cpu, cuda:0, cuda:1, etc.)")
        print(f"    Max Memory:  {model.get('max_memory', 4)} GB")
        # print(f"                 (Maximum RAM/VRAM to use for model layers)")
        print(f"    Load Ends:   {'Yes' if model.get('load_ends') else 'No'}")
        # print(f"                 (Load embedding layer and output head)")
    
    print("\n" + "=" * 60)


def _print_property(label: str, value, description: str):
    """Helper to print a property with its description."""
    print(f"  {label + ':':<18} {value}")



import os
import sys
import time
import unittest

sys.path.insert(0, os.path.dirname(__file__))
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../../src'))

from distributed_state_network.objects.state_packet import StatePacket
from distributed_state_network.objects.hello_packet import HelloPacket

from base import DSNTestBase, spawn_node, remove_node

class TestErrorHandling(DSNTestBase):
    def test_bad_req_data(self):
        """Malformed request data should raise an error"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        try: 
            connector.node.send_http_request(bootstrap.node.my_con(), 1, b'MALFORMED_DATA')
            self.fail("Should throw error for malformed data")
        except Exception as e:
            print(e)

    def test_bad_update_self(self):
        """Node should not handle updates for itself"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        bt_prv_key = bootstrap.node.cred_manager.my_private()
        
        state = StatePacket.create("bootstrap", time.time(), bt_prv_key, {})
        try: 
            bootstrap.node.handle_update(state.to_bytes())
            self.fail("Node should not handle updates for itself")
        except Exception as e:
            print(e)
            self.assertEqual(e.args[0], 406)

    def test_bad_update_unsigned(self):
        """Unsigned state packets should be rejected"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        
        state = StatePacket("connector", time.time(), b'', {})
        try:
            bootstrap.node.handle_update(state.to_bytes())
            self.fail("Should not accept unsigned packets")
        except Exception as e:
            print(e)
            self.assertEqual(e.args[0], 401)

    def test_bad_update_stale(self):
        """Stale state updates should be ignored"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        cn_prv_key = connector.node.cred_manager.my_private()

        time_before = time.time() - 10
        state = StatePacket.create("connector", time.time(), cn_prv_key, {"a": "1"})
        bootstrap.node.handle_update(state.to_bytes())

        state = StatePacket.create("connector", time_before, cn_prv_key, {"a": "2"})
        self.assertFalse(bootstrap.node.handle_update(state.to_bytes()))
    
    def test_bad_hello(self):
        """Stale hello packets should not add invalid peers"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector_0 = spawn_node("connector-0", None, [bootstrap.node.my_con().to_json()])
        connector_0.stop()
        remove_node(connector_0)
        connector_1 = spawn_node("connector-1", "127.0.0.1", [bootstrap.node.my_con().to_json()])
        self.assertEqual(sorted(connector_1.node.peers()), ["bootstrap", "connector-1"])

    def test_bad_packets_hello(self):
        """HelloPacket should reject malformed data"""
        try:
            HelloPacket.from_bytes(b'')
            self.fail("Should throw error on bad parse")
        except Exception as e:
            print(e)

        try:
            HelloPacket.from_bytes(b'Random data')
            self.fail("Should throw error on bad parse")
        except Exception as e:
            print(e)

    def test_bad_packets_state(self):
        """StatePacket should reject malformed data"""
        try:
            StatePacket.from_bytes(b'')
            self.fail("Should throw error on bad parse")
        except Exception as e:
            print(e)

        try:
            StatePacket.from_bytes(b'Random data')
            self.fail("Should throw error on bad parse")
        except Exception as e:
            print(e)


if __name__ == "__main__":
    unittest.main()



import os
import toml
import argparse

from distributed_state_network import DSNodeConfig, DSNodeServer

from language_pipes.config import LpConfig
from language_pipes.util.aes import save_new_aes_key
from language_pipes.commands.initialize import interactive_init
from language_pipes.commands.start import start_wizard
from language_pipes.commands.upgrade import upgrade_lp

from language_pipes.lp import LanguagePipes

VERSION = "0.19.2"

def build_parser():
    parser = argparse.ArgumentParser(
        prog="Language Pipes",
        description="Distribute LLMs across multiple systems"
    )

    parser.add_argument("-v", "--version", action="version", version=VERSION)

    subparsers = parser.add_subparsers(dest="command")

    #Upgrade
    subparsers.add_parser("upgrade", help="Upgrade Language Pipes package")

    # Key Generation
    create_key_parser = subparsers.add_parser("keygen", help="Generate AES key")
    create_key_parser.add_argument("output", nargs='?', help="Output file for AES key (default: network.key)", default="network.key")

    # Initialize
    init = subparsers.add_parser("init", help="Create a new configuration file")
    init.add_argument("output", nargs='?', default="config.toml", help="Output file name to write to (default: config.toml)")

    # run command
    run_parser = subparsers.add_parser("serve", help="Start Language Pipes server")
    run_parser.add_argument("-c", "--config", help="Path to TOML config file")
    run_parser.add_argument("-l", "--logging-level", 
        help="Logging verbosity (Default: INFO)",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"]
    )
    run_parser.add_argument("--openai-port", type=int, help="Open AI server port (Default: none)")
    run_parser.add_argument("--app-data-dir", type=str, help="Application data directory for language pipes (default: ~/.config/language_pipes)")
    run_parser.add_argument("--model-dir", type=str, help="Directory to store model data (default: ~/.cache/language_pipes/models)")
    run_parser.add_argument("--node-id", help="Node ID for the network (Required)")
    run_parser.add_argument("--app-dir", type=str, help="Directory to store data for this application")
    run_parser.add_argument("--peer-port", type=int, help="Port for peer-to-peer network (Default: 5000)")
    run_parser.add_argument("--bootstrap-address", help="Bootstrap node address (e.g. 192.168.1.100)")
    run_parser.add_argument("--bootstrap-port", type=int, help="Bootstrap node port for the network (e.g. 8000)")
    run_parser.add_argument("--max-pipes", type=int, help="Maximum amount of pipes to host")
    run_parser.add_argument("--network-key", type=str, help="AES key to access network (Default: network.key)")
    run_parser.add_argument("--model-validation", help="Whether to validate the model weight hashes when connecting to a pipe.", action="store_true")
    run_parser.add_argument("--hosted-models", nargs="*", metavar="MODEL", 
        help="Hosted models as key=value pairs: id=MODEL,device=DEVICE,memory=GB,load_ends=BOOL (e.g., id=Qwen/Qwen3-1.7B,device=cpu,memory=4,load_ends=false)")
    run_parser.add_argument("--prefill-chunk-size", help="Number of tokens to process for each batch in prefill", type=int)

    return parser

def apply_overrides(data, args):
    # Environment variable mapping
    env_map = {
        "logging_level": os.getenv("LP_LOGGING_LEVEL"),
        "oai_port": os.getenv("LP_OAI_PORT"),
        "app_dir": os.getenv("LP_APP_DIR"),
        "node_id": os.getenv("LP_NODE_ID"),
        "peer_port": os.getenv("LP_PEER_PORT"),
        "network_ip": os.getenv("LP_NETWORK_IP"),
        "bootstrap_address": os.getenv("LP_BOOTSTRAP_ADDRESS"),
        "bootstrap_port": os.getenv("LP_BOOTSTRAP_PORT"),
        "network_key": os.getenv("LP_NETWORK_KEY"),
        "model_validation": os.getenv("LP_MODEL_VALIDATION"),
        "max_pipes": os.getenv("LP_MAX_PIPES"),
        "hosted_models": os.getenv("LP_HOSTED_MODELS"),
        "prefill_chunk_size": os.getenv("LP_PREFILL_CHUNK_SIZE"),
        "model_dir": os.getenv("LP_MODEL_DIR"),
    }

    def precedence(key, arg):
        if arg is not None:
            return arg
        if key in env_map and env_map[key] is not None:
            return env_map[key]
        if key in data:
            return data[key]
        return None

    app_dir_arg = args.app_dir
    if app_dir_arg is None and hasattr(args, "app_data_dir"):
        app_dir_arg = args.app_data_dir

    config = {
        "logging_level": precedence("logging_level", args.logging_level),
        "oai_port": precedence("oai_port", args.openai_port),
        "app_dir": precedence("app_dir", app_dir_arg),
        "node_id": precedence("node_id", args.node_id),
        "peer_port": precedence("peer_port", args.peer_port),
        "network_ip": precedence("network_ip", None),
        "bootstrap_address": precedence("bootstrap_address", args.bootstrap_address),
        "bootstrap_port": precedence("bootstrap_port", args.bootstrap_port),
        "network_key": precedence("network_key", args.network_key),
        "model_validation": precedence("model_validation", args.model_validation),
        "max_pipes": precedence("max_pipes", args.max_pipes),
        "hosted_models": precedence("hosted_models", args.hosted_models),
        "prefill_chunk_size": precedence("prefill_chunk_size", args.prefill_chunk_size),
        "model_dir": precedence("model_dir", args.model_dir),
    }

    if config["peer_port"] is not None:
        config["peer_port"] = int(config["peer_port"])

    if config["hosted_models"] is None:
        print("Error: hosted_models param must be supplied in config")
        exit()

    if config["node_id"] is None:
        print("Error: node_id param is not supplied in config")
        exit()
    
    if config["oai_port"] is not None:
        config["oai_port"] = int(config["oai_port"])
    
    if config["bootstrap_port"] is not None:
        config["bootstrap_port"] = int(config["bootstrap_port"])

    hosted_models = []
    for m in config["hosted_models"]:
        if type(m) is str:
            # Parse Docker-style key=value pairs: id=X,device=Y,memory=Z,load_ends=W
            model_config = {}
            for pair in m.split(","):
                if "=" not in pair:
                    raise ValueError(f"Invalid format '{pair}' in '{m}'. Expected key=value pairs (e.g., id=Qwen/Qwen3-1.7B,device=cpu,memory=4,load_ends=false)")
                key, value = pair.split("=", 1)
                model_config[key.strip()] = value.strip()
            
            # Validate required keys
            required_keys = {"id", "device", "memory"}
            missing = required_keys - set(model_config.keys())
            if missing:
                raise ValueError(f"Missing required keys {missing} in '{m}'")
            
            hosted_models.append({
                "id": model_config["id"],
                "device": model_config["device"],
                "max_memory": float(model_config["memory"]),
                "load_ends": model_config.get("load_ends", "false").lower() == "true"
            })
        else:
            hosted_models.append(m)

    config["hosted_models"] = hosted_models

    return config

def main(argv = None):
    parser = build_parser()
    args = []
    if argv is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(argv)

    # Default to "start" command if no command given
    if args.command is None:
        args.command = "start"
        args.config = "config.toml"
        args.key = "network.key"

    if args.command == "keygen":
        key = save_new_aes_key(args.output)
        print(f"✓ Network key generated: {key}")
        print(f"✓ Network key saved to '{args.output}'")
    elif args.command == "upgrade":
        upgrade_lp()
    elif args.command == "init":
        interactive_init(args.output)
    elif args.command == "start":
        try:
            return start_wizard(apply_overrides, VERSION)
        except KeyboardInterrupt:
            exit()
    elif args.command == "serve":
        data = { }
        if args.config is not None:
            with open(args.config, "r", encoding="utf-8") as f:
                data = toml.load(f)
        data = apply_overrides(data, args)
        
        config = LpConfig.from_dict(data)

        print(config.to_string())

        router_config = DSNodeConfig.from_dict({
            "node_id": data["node_id"],
            "port": data.get("peer_port", 5000),
            "network_ip": data.get("network_ip", None),
            "aes_key": data.get("network_key", None),
            "bootstrap_nodes": [
                {
                    "address": data["bootstrap_address"],
                    "port": data["bootstrap_port"]
                }
            ] if data.get("bootstrap_address") is not None else []
        })

        print(router_config.to_string())

        router = DSNodeServer.start(router_config)

        app = LanguagePipes(config, router)
        return app
    else:
        parser.print_usage()
        exit(1)

if __name__ == "__main__":
    main()



from typing import Optional

from distributed_state_network.objects.endpoint import Endpoint

from distributed_state_network.objects.signed_packet import SignedPacket
from distributed_state_network.util.byte_helper import ByteHelper

class HelloPacket(SignedPacket):
    version: str
    node_id: str
    connection: Endpoint
    ecdsa_public_key: bytes
    detected_address: Optional[str]  # IP address detected by bootstrap node

    def __init__(
        self, 
        version: str, 
        node_id: str, 
        connection: Endpoint,
        ecdsa_public_key: bytes,
        ecdsa_signature: bytes,
        detected_address: Optional[str] = None
    ):
        super().__init__(ecdsa_signature)
        self.version = version
        self.node_id = node_id
        self.connection = connection
        self.ecdsa_public_key = ecdsa_public_key
        self.detected_address = detected_address

    def to_bytes(self, include_signature: bool = True):
        bts = ByteHelper()
        bts.write_string(self.version)
        bts.write_string(self.node_id)
        if self.connection.address is not None:
            bts.write_bytes(self.connection.address.encode())
        else:
            bts.write_bytes(b'')
        bts.write_int(self.connection.port)
        bts.write_bytes(self.ecdsa_public_key)
        if include_signature:
            bts.write_bytes(self.ecdsa_signature)
        # Add detected address (empty string if None)
        bts.write_string(self.detected_address or "")
        
        return bts.get_bytes()

    @staticmethod
    def from_bytes(data: bytes):
        bts = ByteHelper(data)
        version = bts.read_string()
        node_id = bts.read_string()
        addr_bytes = bts.read_bytes()
        address = None
        if addr_bytes != b'':
            address = addr_bytes.decode()
        port = bts.read_int()

        connection = Endpoint(
            address=address,
            port=port
        )
        ecdsa_public_key = bts.read_bytes()
        ecdsa_signature = bts.read_bytes()
        # Read detected address (may be empty string for older packets)
        detected_address = bts.read_string() or None

        if version == '' or node_id == '' or ecdsa_public_key == b'':
            raise Exception(406, "Malformed packet") # Not acceptable

        return HelloPacket(version, node_id, connection, ecdsa_public_key, ecdsa_signature, detected_address)



import os
import sys
import shutil
import unittest
from typing import List, Dict, Optional, Callable

sys.path.append(os.path.join(os.path.dirname(__file__), '../../src'))

from distributed_state_network import DSNodeServer, DSNodeConfig

current_port = 8000
nodes = []

aes_key = DSNodeServer.generate_key()


def spawn_node(
    node_id: str,
    network_ip: Optional[str],
    bootstrap_nodes: List[Dict] = [],
    disconnect_cb: Optional[Callable] = None,
    update_cb: Optional[Callable] = None
):
    global current_port
    current_port += 1
    n = DSNodeServer.start(DSNodeConfig.from_dict({
        "node_id": node_id,
        "port": current_port,
        "network_ip": network_ip,
        "aes_key": aes_key,
        "bootstrap_nodes": bootstrap_nodes
    }), disconnect_cb, update_cb)
    global nodes
    nodes.append(n)
    return n


def get_nodes():
    global nodes
    return nodes


def remove_node(node):
    global nodes
    nodes.remove(node)


class DSNTestBase(unittest.TestCase):
    def tearDown(self):
        global nodes
        for n in nodes:
            n.stop()
        nodes = []

        if os.path.exists('certs'):
            shutil.rmtree('certs')

        if os.path.exists('credentials'):
            shutil.rmtree('credentials')



import os
import sys
import time
import unittest

sys.path.insert(0, os.path.dirname(__file__))

from base import DSNTestBase, spawn_node

class TestStateData(DSNTestBase):
    def test_state_propagation(self):
        """State updates should propagate between nodes"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])

        self.assertEqual(None, bootstrap.node.read_data("connector", "foo"))

        connector.node.update_data("foo", "bar")
        time.sleep(0.5)
        self.assertEqual("bar", bootstrap.node.read_data("connector", "foo"))
        
        bootstrap.node.update_data("bar", "baz")
        time.sleep(0.5)
        self.assertEqual("baz", connector.node.read_data("bootstrap", "bar"))

    def test_multiple_data_updates(self):
        """Multiple data updates should all propagate"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        
        bootstrap.update_data("key1", "value1")
        bootstrap.update_data("key2", "value2")
        bootstrap.update_data("key3", "value3")
        time.sleep(0.5)
        
        self.assertEqual("value1", connector.read_data("bootstrap", "key1"))
        self.assertEqual("value2", connector.read_data("bootstrap", "key2"))
        self.assertEqual("value3", connector.read_data("bootstrap", "key3"))

    def test_send_to_node_success(self):
        """send_to_node should deliver payload to target node"""
        received_data = []
        
        def recv_fn(data: bytes):
            received_data.append(data)
        
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        bootstrap.set_receive_cb(recv_fn)
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])

        payload = b"Hello, world!"
        resp = connector.node.send_to_node("bootstrap", payload)
        self.assertEqual(resp, "OK")
        
        time.sleep(0.5)
        self.assertEqual(len(received_data), 1)
        self.assertEqual(received_data[0], payload)

    def test_send_to_node_wrapper(self):
        """DSNodeServer.send_to_node wrapper should work"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])

        resp = connector.send_to_node("bootstrap", b"Hello via wrapper")
        self.assertEqual(resp, "OK")


if __name__ == "__main__":
    unittest.main()



import json
import random
import logging
from typing import List, Dict, Optional
from language_pipes.network_protocol import StateNetworkNode

from language_pipes.pipes.meta_pipe import MetaPipe
from language_pipes.modeling.meta_model import MetaModel

def aggregate_models(models: List[MetaModel]) -> List[MetaPipe]:
    pipes: List[MetaPipe] = []
    pipe_ids: List[str] = []
    for model in models:
        if model.pipe_id in pipe_ids:
            existing_pipe = [p for p in pipes if p.pipe_id == model.pipe_id][0]
            existing_pipe.segments.append(model)
        else:
            pipes.append(MetaPipe(model.pipe_id, model.model_id, [model]))
            pipe_ids.append(model.pipe_id)

    for pipe in pipes:
        pipe.sort_segments()
    return pipes

class RouterPipes:
    router: StateNetworkNode

    def __init__(self, router: StateNetworkNode):
        self.router = router

    def add_model_to_network(self, model: MetaModel):
        models_string = self.router.read_data(self.router.node_id(), 'models') or '[]'
        current_models = [MetaModel.from_dict(m) for m in json.loads(models_string)]
        current_models.append(model)
        self.router.update_data(
            'models',
            json.dumps([m.to_json() for m in current_models])
        )

    def update_model(self, model: MetaModel):
        models_string = self.router.read_data(self.router.node_id(), 'models') or '[]'
        current_models = [MetaModel.from_dict(m) for m in json.loads(models_string)]
        matching_models = [m for m in current_models if m.process_id == model.process_id]
        if len(matching_models) < 1:
            raise Exception("Could not update model")
        current_models.remove(matching_models[0])
        current_models.append(model)
        self.router.update_data(
            'models',
            json.dumps([m.to_json() for m in current_models])
        )

    def get_pipe_by_pipe_id(self, pipe_id: str) -> Optional[MetaPipe]:
        models: List[MetaModel] = []
        for model in self._all_models():
            if model.pipe_id == pipe_id:
                models.append(model)
        if len(models) == 0:
            return None
        
        return aggregate_models(models)[0]

    def get_pipe_by_model_id(self, model_id: str) -> Optional[MetaPipe]:
        available_pipes: List[MetaPipe] = []
        for p in self.pipes_for_model(model_id, True):
            if not p.is_loading():
                available_pipes.append(p)
        if len(available_pipes) == 0:
            return None

        return random.choice(available_pipes)

    def print_pipes(self, logger: logging.Logger):
        for p in self._network_pipes():
            p.print(logger)

    def _all_models(self) -> List[MetaModel]:
        network_models: Dict[str, List[MetaModel]] = { }
        for peer in self.router.peers():
            peer_models = json.loads(self.router.read_data(peer, 'models') or '[]')
            network_models[peer] = []
            for m in peer_models:
                network_models[peer].append(MetaModel.from_dict(m))
        models: List[MetaModel] = []
        for key in network_models.keys():
            models.extend(network_models[key])
        return models

    def pipes_for_model(self, model_id: str, find_completed: bool) -> List[MetaPipe]:
        models: List[MetaModel] = []
        for model in self._all_models():
            if model.model_id != model_id:
                continue
            models.append(model)

        return [pipe for pipe in aggregate_models(models) if pipe.is_complete() == find_completed]

    def _network_pipes(self) -> List[MetaPipe]:
        return aggregate_models(self._all_models())

    def get_models(self) -> List[str]:
        model_ids = []
        for pipe in self._network_pipes():
            if pipe.is_complete() and pipe.model_id not in model_ids:
                model_ids.append(pipe.model_id)
        return model_ids
                


import hashlib
from typing import Tuple

from ecdsa import SigningKey, VerifyingKey, SECP256k1

def generate_key_pair() -> Tuple[bytes, bytes]:
    private_key = SigningKey.generate(curve=SECP256k1)
    public_key = private_key.get_verifying_key()
    if public_key is None:
        raise Exception("Error generating keys")
    return public_key.to_string(), private_key.to_string()

def sign_message(private_key_bytes: bytes, message: bytes) -> bytes:
    private_key = SigningKey.from_string(private_key_bytes, curve=SECP256k1)
    message_hash = hashlib.sha256(message).digest()
    return private_key.sign(message_hash)

def verify_signature(public_key: bytes, message: bytes, signature: bytes):
    public_key_obj = VerifyingKey.from_string(public_key, curve=SECP256k1)
    message_hash = hashlib.sha256(message).digest()
    try:
        return public_key_obj.verify(signature, message_hash)
    except Exception:
        return False



# Allows testing of cli by using `python src/langugage_pipes/test_cli.py` when at the repository base directory

import os
import sys
import pathlib
cd = pathlib.Path().resolve()
sys.path.append(os.path.join(cd, 'src'))

from language_pipes.cli import main

main(sys.argv[1:])



from io import BytesIO
import struct


def int_to_bytes(i: int) -> bytes:
    return i.to_bytes(4, "little", signed=False)


def bytes_to_int(b: bytes) -> int:
    return int.from_bytes(b, "little", signed=False)


def float_to_bytes(f: float) -> bytes:
    return struct.pack(">d", f)


def bytes_to_float(b: bytes) -> float:
    return struct.unpack(">d", b)[0]


class ByteHelper:
    def __init__(self, data: bytes | None = None):
        self.bts = BytesIO(data) if data is not None else BytesIO()

    def write_string(self, s: str):
        encoded = s.encode("utf-8")
        self.write_bytes(encoded)

    def write_int(self, i: int):
        self.bts.write(int_to_bytes(i))

    def write_float(self, f: float):
        self.bts.write(float_to_bytes(f))

    def write_bytes(self, b: bytes):
        self.bts.write(int_to_bytes(len(b)))
        self.bts.write(b)

    def read_string(self):
        return self.read_bytes().decode("utf-8")

    def read_int(self):
        return bytes_to_int(self.bts.read(4))

    def read_float(self):
        return bytes_to_float(self.bts.read(8))

    def read_bytes(self):
        length = bytes_to_int(self.bts.read(4))
        return self.bts.read(length)

    def get_bytes(self):
        return self.bts.getvalue()



import os
import toml
from time import sleep
from pathlib import Path

from distributed_state_network import DSNodeServer, DSNodeConfig

from language_pipes.config import LpConfig, default_config_dir, default_model_dir
from language_pipes.commands.initialize import interactive_init
from language_pipes.commands.edit import edit_config
from language_pipes.commands.view import view_config
from language_pipes.lp import LanguagePipes
from language_pipes.util.user_prompts import prompt, prompt_number_choice, select_config, get_config_files, show_banner
from language_pipes.util import sanitize_file_name

def start_server(app_dir: str, config_path: str):
    show_banner("Starting server")
    
    with open(config_path, "r", encoding="utf-8") as f:
        data = toml.load(f)

    data["app_dir"] = app_dir
    
    config = LpConfig.from_dict(data)

    try:
        router = DSNodeServer.start(DSNodeConfig.from_dict({
            "node_id": data.get("node_id"),
            "port": data.get("peer_port", 5000),
            "network_ip": data.get("network_ip", None),
            "credential_dir": str(Path(app_dir) / "credentials"),
            "aes_key_file": data.get("network_key", None),
            "bootstrap_nodes": [
                {
                    "address": data["bootstrap_address"],
                    "port": data["bootstrap_port"]
                }
            ] if data.get("bootstrap_address") is not None else []
        }))
        
        LanguagePipes(config, router)
        while True:
            sleep(60)
    except KeyboardInterrupt:
        exit()
    except Exception as e:
        print(e)

def new_config(app_dir: str):
    raw_name = prompt("Name of new configuration", required=True)
    if raw_name is None:
        return
    file_name = sanitize_file_name(raw_name)
    if not file_name:
        print("Invalid file name")
        return
    
    file_path = str(Path(app_dir) / "configs" / file_name)

    interactive_init(file_path)

def delete_config(app_dir: str):
    config_path = select_config(app_dir)
    if config_path is not None:
        os.remove(config_path)
        print("Configuration deleted")

def modify_config(app_dir: str):
    config_path = select_config(app_dir)
    if config_path:
        edit_config(config_path)

def start_wizard(apply_overrides, version: str):
    print(f"""         
==============================================================================

 | |                                              |  __ (_)                
 | |     __ _ _ __   __ _ _   _  __ _  __ _  ___  | |__) | _ __   ___  ___ 
 | |    / _` | '_ \ / _` | | | |/ _` |/ _` |/ _ \ |  ___/ | '_ \ / _ \/ __|
 | |___| (_| | | | | (_| | |_| | (_| | (_| |  __/ | |   | | |_) |  __/\__ \\
 |______\__,_|_| |_|\__, |\__,_|\__,_|\__, |\___| |_|   |_| .__/ \___||___/
                     __/ |             __/ |              | |              
                    |___/             |___/               |_|      
Version: {version}
==============================================================================

- Made with <3 by Erin
""")

    app_dir = default_config_dir()
    model_dir = default_model_dir()
    
    if not os.path.exists(app_dir):
        Path(app_dir).mkdir(parents=True)
        print(f"Created directory: {app_dir}")
    
    config_dir = str(Path(app_dir) / "configs")
    if not os.path.exists(config_dir):
        Path(config_dir).mkdir(parents=True)

    if not os.path.exists(model_dir):
        Path(model_dir).mkdir(parents=True)

    while True: # TODO Add model list        
        main_menu_cmd = prompt_number_choice("Main Menu", [
            "View Config",
            "Load Config",
            "Create Config",
            "Edit Config",
            "Delete Config"
        ])

        print("\n" + ("=" * 50) + "\n")

        match main_menu_cmd:
            case "Load Config":
                if len(get_config_files(config_dir)) == 0:
                    print("No configs found...\n\n")
                    continue
                config_path = select_config(app_dir)
                if config_path is not None:
                    view_config(config_path)
                    start_server(app_dir, config_path)
            case "View Config":
                if len(get_config_files(config_dir)) == 0:
                    print("No configs found...\n\n")
                    continue
                config_path = select_config(app_dir)
                if config_path is not None:
                    view_config(config_path)
            case "Create Config":
                new_config(app_dir)
            case "Edit Config":
                if len(get_config_files(config_dir)) == 0:
                    print("No configs found...\n\n")
                    continue
                modify_config(app_dir)
            case "Delete Config":
                if len(get_config_files(config_dir)) == 0:
                    print("No configs found...\n\n")
                    continue
                delete_config(app_dir)

        print("\n" + ("=" * 50) + "\n")



import os
import socket
import toml

from language_pipes.util.aes import generate_aes_key
from language_pipes.commands.view import view_config
from language_pipes.util.user_prompts import prompt, prompt_bool, prompt_choice, prompt_float, prompt_int, prompt_model_id, show_banner

def interactive_init(output_path: str):
    show_banner("Configuration Setup")

    config = { }

    print("--- Required Settings ---\n")

    print("The node ID is a unique name that identifies this computer on the")
    print("Language Pipes network. Other nodes will use this to route jobs to this machine.\n")
    config["node_id"] = prompt(
        "    Node ID",
        default=socket.gethostname(),
        required=True
    )

    print("\n--- Model Configuration ---\n")
    print("You need to specify at least one model to host.")
    print("Models are automatically downloaded from HuggingFace by their ID")
    print("if they do not exist in the model folder already.\n")

    hosted_models = []
    try:
        while True:
            print(f"    Model #{len(hosted_models) + 1}:")
            
            print("    Select a locally available model or enter a HuggingFace model ID")
            model_id = prompt_model_id(
                "    Model ID",
                required=True
            )
            
            print("\n    Select the compute device for this model. Use 'cpu' for CPU-only,")
            print("    or 'cuda:0', 'cuda:1', etc. for specific GPUs.")
            device = prompt(
                "    Device",
                default="cpu",
                required=True
            )
            
            print("\n    Specify the maximum RAM/VRAM (in GB) this node should use for the model.")
            print("    The model layers will be loaded until this limit is reached.")
            max_memory = prompt_float(
                "    Max memory (GB)",
                required=True
            )
            
            print("\n    The 'ends' of a model are the embedding layer (input) and the output head.")
            print("    For the privacy protection to work correctly the requesting user must control an end node")
            print("    It is possible for many nodes on the network to be an end node so long as all the layers are loaded on the network")
            load_ends = prompt_bool(
                "    Load embedding/output layers",
                required=True
            )
            
            hosted_models.append({
                "id": model_id,
                "device": device,
                "max_memory": max_memory,
                "load_ends": load_ends
            })
            
            print()
            if not prompt_bool("Add another model?", required=True):
                break
    except KeyboardInterrupt:
        pass

    config["hosted_models"] = hosted_models

    # === API Server ===
    print("\n--- API Server ---\n")

    print("Language Pipes can expose an OpenAI-compatible HTTP API, allowing you to")
    print("use standard OpenAI client libraries (Python, JavaScript, curl, etc.)")
    print("to interact with your distributed model.\n")
    if prompt_bool("Enable OpenAI-compatible API server?", default=True):
        print("\n  Choose a port for the API server. Clients will connect to")
        print("  http://<ip-address>:<port>/v1/chat/completions")
        config["oai_port"] = prompt_int(
            "  API port",
            default=8000,
            required=True
        )

    # === Network Configuration ===
    print("\n--- Network Configuration ---\n")

    print("Language Pipes uses a peer-to-peer network to coordinate between nodes.")
    print("The first node starts fresh; additional nodes connect to an existing node.\n")
    is_first_node = prompt_bool("Is this the first node in the network?", required=True)

    if not is_first_node:
        print("\n  Enter the IP address of an existing node on the network.")
        print("  This node will connect to it to join the distributed network.")
        config["bootstrap_address"] = prompt(
            "  Bootstrap node IP",
            required=True
        )
        print("\n  Enter the peer port of the bootstrap node (default is 5000).")
        config["bootstrap_port"] = prompt_int(
            "  Bootstrap node port",
            default=5000,
            required=True
        )

    print("\nThe peer port is used for network coordination and discovery.")
    print("Other nodes will connect to this port to join the network.")
    config["peer_port"] = prompt_int(
        "Peer port",
        default=5000
    )

    if is_first_node:
        print("\nThe network IP is the address other nodes will use to connect to this node.")
        config["network_ip"] = prompt(
            "Network IP",
            required=True
        )
    else:
        del config["network_ip"]

    print("\nThe network key is an AES encryption key shared by all nodes.")
    print("It encrypts communication and prevents unauthorized access.")
    print("There will be no encryption between nodes if the default value is selected.")
    print("")
    encrypt_traffic = prompt_bool("Encrypt network traffic", default=False)
    if encrypt_traffic:
        config["network_key"] = prompt(
            "Network key",
            default="Generate new key"
        )

        if config["network_key"] == "Generate new key":
            key = generate_aes_key().hex()
            config["network_key"] = key
            print(f"Generated new key: {key}")
            print("Note: Save this key somewhere and supply it to other nodes on the network")
    else:
        if "network_key" in config:
            del config["network_key"]

    # === Advanced Options ===
    print("\n--- Advanced Options ---\n")

    print("Advanced options include logging verbosity, security features, and limits.")
    if prompt_bool("Configure advanced options?", default=False):
        print("\n  Controls how much information is printed to the console.")
        print("  DEBUG shows everything, ERROR shows only critical issues.")
        config["logging_level"] = prompt_choice(
            "  Logging level",
            ["DEBUG", "INFO", "WARNING", "ERROR"],
            default="INFO"
        )
        
        print("\n  Limits how many model 'pipes' (distributed model instances)")
        print("  this node will participate in simultaneously.")
        config["max_pipes"] = prompt_int(
            "  Max pipes",
            default=1
        )
        
        print("\n  When enabled, nodes verify that model weight hashes match")
        print("  to ensure all nodes are running the exact same model.")
        config["model_validation"] = prompt_bool(
            "  Validate model hashes?", 
            required=True
        )
        
    # === Write Config File ===
    print("\n" + "=" * 50)
    print("  Configuration Summary")
    print("=" * 50 + "\n")

    # Build clean config (remove None values)
    clean_config = {k: v for k, v in config.items() if v is not None and k != "hosted_models"}
    clean_config["hosted_models"] = config["hosted_models"]

    # Check if file exists
    if os.path.exists(output_path):
        if not prompt_bool(f"  '{output_path}' already exists. Overwrite?", default=False):
            print("Aborted.")
            return
    
    with open(output_path, 'w', encoding='utf-8') as f:
        toml.dump(clean_config, f)

    view_config(output_path)
    
    print("\n✓ Configuration saved")



import io
import re
import os
import shutil
import struct
from pathlib import Path
import json
import base64
import ctypes
import subprocess
from uuid import UUID
from hashlib import sha256
from threading import Thread
from typing import Optional

import numpy as np
import torch

def uuid_to_bytes(uid: str) -> bytes:
    return UUID(hex=uid).bytes

def bytes_to_uuid(b: bytes) -> str:
    return str(UUID(bytes=b))

def int_to_bytes(i: int) -> bytes:
    return i.to_bytes(4, 'little', signed=False)

def bytes_to_int(b: bytes) -> int:
    return int.from_bytes(b, 'little', signed=False)

# Fast tensor serialization: dtype code (1 byte) + ndim (1 byte) + shape (4 bytes each) + raw data
_DTYPE_TO_CODE = {
    torch.float32: 0, torch.float64: 1, torch.float16: 2,
    torch.int32: 3, torch.int64: 4, torch.int16: 5, torch.int8: 6,
    torch.uint8: 7, torch.bool: 8, torch.bfloat16: 9,
}
_CODE_TO_DTYPE = {v: k for k, v in _DTYPE_TO_CODE.items()}

def tensor_to_bytes(t: torch.Tensor | None) -> bytes:
    if t is None:
        return b''
    t_cpu = t.detach().cpu().contiguous()
    dtype_code = _DTYPE_TO_CODE.get(t_cpu.dtype)
    if dtype_code is None:
        # Fallback for unsupported dtypes
        bts = io.BytesIO()
        torch.save(t_cpu, bts)
        return b'\xff' + bts.getvalue()
    
    shape = t_cpu.shape
    ndim = len(shape)
    header = struct.pack('<BB' + 'I' * ndim, dtype_code, ndim, *shape)
    
    # bfloat16 doesn't have numpy support, so view as int16
    if t_cpu.dtype == torch.bfloat16:
        data = t_cpu.view(torch.int16).numpy().tobytes()
    else:
        data = t_cpu.numpy().tobytes()
    
    return header + data

def bytes_to_tensor(b: bytes) -> torch.Tensor | None:
    if b == b'':
        return None
    
    if b[0] == 0xff:
        # Fallback format
        return torch.load(io.BytesIO(b[1:]), weights_only=True)
    
    dtype_code = b[0]
    ndim = b[1]
    shape = struct.unpack('<' + 'I' * ndim, b[2:2 + 4*ndim])
    dtype = _CODE_TO_DTYPE[dtype_code]
    data = b[2 + 4*ndim:]
    
    if dtype == torch.bfloat16:
        arr = np.frombuffer(data, dtype=np.int16).reshape(shape).copy()
        return torch.from_numpy(arr).view(torch.bfloat16)
    else:
        np_dtype = torch.zeros(1, dtype=dtype).numpy().dtype
        arr = np.frombuffer(data, dtype=np_dtype).reshape(shape).copy()
        return torch.from_numpy(arr)

def get_tensor_byte_string(t: torch.Tensor) -> str:
    bts = tensor_to_bytes(t)
    return base64.b64encode(bts).decode('utf-8')

def get_hash(data: str) -> str:
    hash = sha256(data.encode())
    return hash.hexdigest()

def get_dict_hash(data: dict):
    return get_hash(json.dumps(data))

def size_of_tensor(t: torch.Tensor):
    return t.element_size() * t.nelement()

def tensor_hash(t: torch.Tensor) -> str:
    return get_hash(get_tensor_byte_string(t))

def stop_thread(thread: Thread):
    thread_id = thread.ident
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id,
            ctypes.py_object(SystemExit))
    if res > 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
        print('Exception raise failure')

def clone_model(model_id: str, model_dir: str):
    repo_url = f"https://huggingface.co/{model_id}"
    clone_dir = f"{model_dir}/data"

    if not os.path.exists(clone_dir):
        Path(clone_dir).mkdir(parents=True)
    try:
        subprocess.run(["git", "clone", repo_url, clone_dir])
        subprocess.run(["git", "lfs", "install"], cwd=clone_dir, check=True)
        subprocess.run(["git", "lfs", "pull"], cwd=clone_dir, check=True)
    except Exception as e:
        if os.path.exists(model_dir):
            shutil.rmtree(model_dir)
        print(e)
        print("Git LFS Error occurred: please ensure that git-lfs is installed")
        exit()


def sanitize_file_name(raw_name: str):
    return re.sub(r'[<>:"/\\|?*\x00-\x1f]', '_', raw_name).strip().strip('.') + ".toml"

def raise_exception(logger, msg: str):
    logger.exception(msg)
    raise Exception(msg)

def maybeTo(t: Optional[torch.Tensor], device: str) -> Optional[torch.Tensor]:
    if t is None:
        return None
    if str(t.device) == device:
        return t.detach()
    return t.detach().to(device)
